import os
import sys
from sys import platform

''' Detect whether we are in klayout's generic scripting interface
    Since pya cannot be imported from outside, try that.
'''
try:
    import pya
    _isGSI = True
except ImportError:
    _isGSI = False
isGSI = lambda: _isGSI


''' Klayout can run as a window or in batch mode on command line.
    If it launches in batch (-b) or database-only (-zz) mode, then main_window is None.

    If however it runs in non-GUI mode (-z), it is not None; however,
    it has not been given a title yet. (see layApplication.cc)
'''
if isGSI():
    import pya
    main = pya.Application.instance().main_window()
    if main is not None:
        _isGUI = True
    else:
        _isGUI = False
else:
    _isGUI = False
isGUI = lambda: _isGUI


def klayout_home():
    ''' Figure out the klayout configuration directory.
        Uses exactly the same logic as used in the klayout source.
    '''
    ly_home = os.environ.get('KLAYOUT_HOME', '')
    if ly_home == '':
        ly_home = os.path.join(os.path.expanduser('~'), 'KLayout' if is_windows() else '.klayout')
    if not os.path.exists(ly_home):
        print('The KLayout config directory was not found. KLayout might not be installed.')
        print('Creating it just in case that solves things')
        os.mkdir(ly_home)
        # raise FileNotFoundError('The KLayout config directory was not found. KLayout might not be installed.')
    return ly_home


def klayout_version():
    return '0.25.3'  # TODO: make this not hard coded


def is_windows():
    if platform == "linux" or platform == "linux2":
        return False
    elif platform == "darwin":
        return False
    elif platform == "win32":
        return True
    else:
        raise ValueError('Unrecognized operating system: {}'.format(platform))



### The important stuff happens below ###


''' Determine what we will use for pya.
    Do not override it with the standalone in sys.modules yet
    This statement:
        from lygadgets import pya
    will not error even if you are not in klayout GSI and don't have the klayout.db standalone
'''
if isGSI():
    import pya
    using_standalone = False
else:
    try:
        import klayout.db as pya
        using_standalone = True
    except ImportError as err:
        pya = None
        using_standalone = False


''' Spoof a whole bunch of stuff related to pya GUI.

    The problem this is trying to solve is when layout modules try to import GUI things and call them.
    When you run them in standalone, either those QElements don't exist, or they error for whatever reason.
    Nevertheless, you want to import that module to access its other functionality

    The desired behavior of these GUI things is pretty simple: don't error, do nothing.
    At the same time, we'd like the ability to override our override
    in case there are one or two important things, like instance.version() returning a string
'''

class NS_Catcher(type):
    ''' All this does is override the pya namespace with this class '''
    def __init__(cls, name, bases, dct):
        setattr(pya, name, cls)
        super().__init__(name, bases, dct)

    def __getattr__(cls, attr):
        return PhonyClass()


class PhonyClass(metaclass=NS_Catcher):
    ''' It only ever gives instances of PhonyClass when called or as attributes.
        It is good for stifling those long chained calls like::

            pya.QFormBuilder().load(ui_file, pya.Application.instance().main_window()).findChild('ok').clicked(self.ok)

        That call will do nothing of course, but it also won't error.
    '''
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return PhonyClass()

    def __getattr__(self, attr):
        return PhonyClass()

    def __setattr__(self, attr, value):
        pass


def patch_environment():
    ''' Run this function, then your old scripts for klayout will work with the standalone.
        That's the idea anyways.

        Now we start changing things in the environment. Afterwards, this statement:
            import pya
        might give you regular pya or klayout.db, depending.

        GUI and Application things are added and/or spoofed.
    '''
    if using_standalone:
        sys.modules['pya'] = pya

    if not isGUI():
        class QMessageBox(PhonyClass): pass

        class QMessageBox_StandardButton(PhonyClass): pass

        class QFile(PhonyClass): pass

        class QIODevice(PhonyClass): pass

        class QFormBuilder(PhonyClass): pass

        class MessageBox(PhonyClass): pass

    if not isGSI():
        class PhonyInstance(PhonyClass):
            ''' This has to return a string sometimes '''
            def application_data_path(self):
                return klayout_home()

            def version(self):
                return klayout_version()

        class Application(PhonyClass):
            instance = PhonyInstance




from setuptools import setup


def readme():
    with open('README.md', 'r') as fx:
      return fx.read()


setup(name='lygadgets',
      version='0.0.15',
      description='Tools to make klayout and python work better together',
      long_description=readme(),
      author='Alex Tait',
      author_email='alexander.tait@nist.gov',
      license='MIT',
      packages=['lygadgets'],
      install_requires=[],
      entry_points={'console_scripts': ['lygadget_link=lygadgets.command_line:cm_link_any']},
      )



from __future__ import print_function
try:  # python 2 does not have FileNotFoundError. Check it
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

import os
import sys
import subprocess
from importlib import import_module

from lygadgets.messaging import message, message_loud
from lygadgets.environment import is_windows

pypackage = None

# klayout's version
def get_klayout_version():
    klayouts_pypackage = import_module(pypackage)
    try:
        return klayouts_pypackage.__version__
    except AttributeError as err:
        err.args = ('The package {} does not define a __version__ in its __init__.py'.format(pypackage), ) + err.args[:1]
        raise


_system_python = None
def system_python():
    ''' Fix the PATH to prioritize /usr/local/bin

        This is no longer windows compatible.
        If we are running in windows, it always raises a
    '''
    if is_windows():
        raise RuntimeError('finding system_python on Windows not supported.')
    global _system_python
    if _system_python is None:
        # Fix the PATH to prioritize /usr/local/bin
        # This is no longer windows compatible

        # def shell_call(commands, env=None, shell=False):
        #     if env is None:
        #         env = os.environ
        #     retraw = subprocess.check_output(commands, env=env, shell=shell)
        #     return retraw.decode().strip("'").strip()

        def is_py3(pyexecutable):
            retraw = subprocess.call([pyexecutable, '-c', '"import sys; assert sys.version_info >= (3, 1)"'])
            return retraw == 0

        # def interactive_bash_call(bash_commands):
        #     full_command = ['/bin/bash', '-i', '-c']
        #     full_command.append('"{}"'.format('; '.join(bash_commands)))
        #     return shell_call(full_command, env=None, shell=False)

        # def as_true_user(bash_commands):
        #     full_bash_commands = []
        #     config_files = ['.bash_profile', '.bashrc', '.bash_aliases']
        #     user_home = os.environ['HOME']
        #     for cf in config_files:
        #         full_file = os.path.join(user_home, cf)
        #         full_bash_commands.append('source ' + full_file)
        #     full_bash_commands.extend(bash_commands)
        #     return interactive_bash_call(full_bash_commands)

        # are we dealing with an alias?
        # THIS CAUSES HANG. REMOVING
        # ret = as_true_user(['type python'])
        # if 'aliased to' in ret:
        #     raise FileNotFoundError('You have aliased python. This is not supported. Please change it to a symlink')
        #     # start_ind = ret.find('aliased to')
        #     # resolved_proto = ret[start_ind+11:]
        #     # resolved = resolved_proto.strip("`'").strip()
        # else:
        #     resolved = 'python'
        resolved = 'python'

        for search_path in ['/usr/local/bin', '/usr/local/python', '/usr/local/opt']:
            my_env = os.environ.copy()
            my_env['PATH'] = search_path + ':' + my_env['PATH']
            retraw = subprocess.check_output(['which', resolved], env=my_env)
            pyexecutable = retraw.decode().strip("'").strip()
            if is_py3(pyexecutable):
                _system_python = pyexecutable
                break
        else:
            raise FileNotFoundError('python version 3 was not found in any of the normal places. Make sure it is installed.')
    return _system_python


# System version
def get_system_version():
    check_program = ('import {}; '.format(pypackage) +
                     'print({}.__version__)'.format(pypackage))
    try:
        retraw = subprocess.check_output([system_python(), '-c', check_program])
        all_talking = retraw.decode().strip("'").strip('\n')
        printed_lines = all_talking.split('\n')
        return printed_lines[-1]
    except AttributeError as err:
        err.args = ('The package {} does not define a __version__ in its __init__.py'.format(pypackage), ) + err.args[:1]
        raise
    except (ImportError, ModuleNotFoundError, subprocess.CalledProcessError, RuntimeError):
        # It's not installed at all
        return '-1'


# Source version
def get_source_version(pypackage_dir):
    version_file = os.path.join(pypackage_dir, pypackage, '__init__.py')
    if sys.version_info.major >= 3:
        import importlib.util
        spec = importlib.util.spec_from_file_location(pypackage, version_file)
        source_version_module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(source_version_module)
        except Exception as err:
            message_loud(('Error loading {}!\n\n{}\n\n'.format(pypackage, err) +
                        'Get the traceback by launching klayout from command line'))
            raise
    else:
        import imp
        source_version_module = imp.load_source(pypackage, version_file)
    try:
        return source_version_module.__version__
    except AttributeError as err:
        err.args = ('The package {} does not define a __version__ in its __init__.py'.format(pypackage), ) + err.args[:1]
        raise


def install_from_source(source_dir):
    install_call = [system_python(), 'setup.py', 'install']
    try:
        subprocess.check_call(install_call, cwd=source_dir)
        message('Success')
    except Exception as err:
        message_loud(('Error installing {}!\n\n{}\n\n'.format(pypackage, err) +
                      'Get the traceback by launching klayout from command line (verified)'))
        import traceback
        message_loud(traceback.extract_stack())
        raise


def export_to_system(pypackage_name, lypackage_dir):
    global pypackage
    pypackage = pypackage_name
    pypackage_dir = os.path.join(lypackage_dir, 'python')
    source_version = get_source_version(pypackage_dir)

    if source_version == get_klayout_version() and source_version == get_system_version():
        message('{} {} already installed'.format(pypackage, source_version))
    else:
        message(('{} ({}) '.format(pypackage, get_klayout_version()) +
                 'is desynchronized from source ({})'.format(source_version)))
        install_from_source(pypackage_dir)




from __future__ import print_function
try:  # python 2 does not have FileNotFoundError. Check it
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

import os
import subprocess
from setuptools.command.install import install
from importlib import import_module
from shutil import rmtree, copytree
from types import ModuleType

from lygadgets.markup import xml_to_dict
from lygadgets.environment import klayout_home, is_windows


def symlink_windows(source, destination):
    # API based. The easy way
    try:
        os.symlink(source, destination)
        return
    except AttributeError:
        pass

    # Command line shell
    try:
        retval = subprocess.call(['ln', '-s', source, destination])
        assert retval == 0
        return
    except (subprocess.CalledProcessError, WindowsError, AssertionError):
        pass

    # Big magic with windows-specific package
    # From https://stackoverflow.com/questions/1447575/symlinks-on-windows
    try:
        import ctypes
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 0
        if source is not None and os.path.isdir(source):
            flags = 1
        if csl(destination, source, flags) == 0:
            raise ctypes.WinError('You do not have the permission to create a symbolic link.\nTry running as administrator')
        return
    except WindowsError:
        pass

    # Otherwise it didn't work
    raise WindowsError('Failed to find a way to link to a file in windows')


def lypackage_name(source):
    try:
        with open(os.path.join(source, 'grain.xml')) as grain:
            return xml_to_dict(grain.read())['salt-grain']['name']
    except FileNotFoundError as err:
        err.args = ((source +
                     ' does not appear to be a klayout package.' +
                     'It must have a grain.xml file.\n' +
                     str(err.args[0])), ) + err.args[1:]
        raise
    except KeyError as err:
        raise FileNotFoundError((source +
                                 ' : grain.xml does not define a "name" property.'))


def is_lypackage(source):
    if not os.path.isdir(source):
        return False
    try:
        lypackage_name(source)
        return True
    except (FileNotFoundError, KeyError):
        return False


def is_pypackage(source):
    if not os.path.isdir(source):
        return False
    return os.path.exists(os.path.join(source, '__init__.py'))


def is_pymodule(source):
    if not os.path.isfile(source):
        return False
    return os.path.splitext(source)[1] == '.py'


def is_lytech(source):
    for file_obj in os.listdir(source):
        if file_obj.endswith('.lyt'):
            return True
    else:
        return False


def module_from_str(module):
    if type(module) is ModuleType:
        return module
    elif type(module) is str:
        return import_module(module)
    else:
        raise TypeError('Argument must either be a module or a string')


def is_installed_python(module):
    try:
        module_from_str(module)
        return True
    except:
        return False


def srcdir_from_any(source):
    if os.path.exists(source):
        return source
    elif is_installed_python(source):
        module = module_from_str(source)
        return module.__path__[0]
    else:
        raise FileNotFoundError('{} does not exist'.format(source))


def dest_from_srcdir(source):
    if is_lypackage(source):
        link_name = lypackage_name(source)
        link_dir = os.path.join(klayout_home(), 'salt')
    elif is_pypackage(source) or is_pymodule(source):
        link_dir = os.path.join(klayout_home(), 'python')
        link_name = os.path.splitext(os.path.basename(source))[0]
    elif is_lytech(source):
        link_dir = os.path.join(klayout_home(), 'tech')
        link_name = os.path.splitext(os.path.basename(source))[0]

    if not os.path.exists(link_dir):
        os.mkdir(link_dir)
    link = os.path.join(link_dir, link_name)
    return link


def link_any(any_source, overwrite=False, hard_copy=False):
    ''' Directories take precedence over installed python module

        Platform independent.

        Always overwrites existing symbolic links.

        Returns the full paths of source and destination if the link was created, otherwise None for both

    '''
    any_source = os.path.realpath(any_source)
    src = srcdir_from_any(any_source)
    dest = dest_from_srcdir(src)

    if src == dest:
        # Prevent circular reference
        return None, None
    if os.path.islink(dest):
        os.remove(dest)
    if os.path.exists(dest):
        if overwrite:
            rmtree(dest)
        else:
            return None, None

    if hard_copy:
        copytree(src, dest)
    else:
        if not is_windows():
            os.symlink(src, dest)
        else:
            symlink_windows(src, dest)

    return src, dest


def postinstall_hook(source):
    ''' Generates a class that subclasses install.
        It attempts to link the lypackage_dir into salt.

        It can be run by setup.py with the argument::

            cmdclass={'install': post_install_factory(some_directory)},

        If you use pip, this sript will be blocked. So this method is no longer recommended.
        Instead, inform the user that they should run

            lygadgets_link yourpackage

        on the command line, after installing via pip
    '''
    class PostInstall(install):
        def run(self):
            install.run(self)
            try:
                the_link = link_any(source)
                if the_link is not None:
                    print('\nAutoinstall into klayout succeeded!\n{}\n'.format(the_link))
                else:
                    print('Folder already present. Not overwriting.')
            except Exception as err:
                print('Autoinstall into klayout failed with the following\n')
                print(err)
                print('\nYou must perform a manual install as described in a README.')
    return PostInstall
